/******************* Icon styling *******************/

/*
Instructions on editing the icons used for the toolbar.

	1. Go to https://icomoon.io/app/#/select
	2. In the top left corner press import icons and select all the icons in
		the openmdao/visualization/n2_viewer/assets/toolbar_graphics/icomoon_icons/SVG folder
	4. Depending on whether you want to remove, replace, or edit an icon:
		- If you want to remove an icon, in the icomoon toolbar there should be a trash icon.
			Click that and then click the icon you wish to remove. Click the Select button in the icomoon toolbar
			if you are finished removing icons to go back to selection mode.
		- If you want to make a new icon, you need to use either Sketch or Adobe Illustrator, or any other SVG editing
			tool and make a new SVG file. Then using the hamburger menu on the icomoon site, select Import to Set and
			select your new SVG file.
		- If you want to modify an icon, edit the existing SVG file located
			in openmdao/visualization/n2_viewer/assets/toolbar_graphics/icomoon_icons/SVG. Be sure to keep the same
			name of the file
			Then on the icomoon site, press on the hamburger menu on the right side and select Import to Set.
			Select the modified SVG file.
			A dialog will popup asking Replace existing icons?. Click the Replace button.
			This process will overwrite the existing icon on the icomoon site with the imported one.
	5. Once you have made all the changes you wish to make, you need to select all the icons by using the hamburger
		menu and clicking Select All, then in the bottom right corner press Generate Font.
		Then press Download in the bottom right
	6. Save the zip file and unzip it
	7. If you added or removed any icons, you need to update the icomoon section of this css file. See the next
		section. Open up the style.css file downloaded from icomoon. Copy everything except the @font-face part of that
		file. Paste that into the contents of the icomoon section of this file replacing it.
	8. Replace the existing openmdao/visualization/n2_viewer/style/icomoon.woff file with the one in the zip file
		downloaded from icomoon.
	9. If you want the icons in a format other than SVG, from the icomoon Web app page, click Generate SVG & More.
		From the page that is then displayed, click the gear icon in lower left next to the Download button. In the
		window that comes up, select the formats you want and close the window. Click the Download button. 
*/

/* Beginning of section is copied from the icomoon style.css file that is generated by icomoon and is contained in the zip
	file that is downloaded from icomoon
 */
[class^='icon-'],
[class*=' icon-'] {
	/* use !important to prevent issues with browser extensions that change fonts */
	font-family: 'icomoon' !important;
	speak: none;
	font-style: normal;
	font-weight: normal;
	font-variant: normal;
	text-transform: none;
	line-height: 1;

	/* Better Font Rendering =========== */
	-webkit-font-smoothing: antialiased;
	-moz-osx-font-smoothing: grayscale;
}

.icon-question-circle:before {
	content: '\e90d';
}

.icon-bezier-curve:before {
	content: '\e908';
}

.icon-caret-right:before {
	content: '\e909';
}

.icon-chevron-left:before {
	content: '\e90a';
}

.icon-info-circle:before {
	content: '\e90b';
}

.icon-key:before {
	content: '\e90c';
}

.icon-save:before {
	content: '\e90e';
}

.icon-forward:before {
	content: '\e906';
}

.icon-forward {
	font-size: 13px !important;
}

.icon-backward:before {
	content: '\e907';
}

.icon-backward {
	font-size: 13px !important;
}

.icon-collapse-depth:before {
	content: '\e905';
}

.icon-hide-connections:before {
	content: '\e900';
}

.icon-hide-connections-2:before {
	content: '\e900';
}

.icon-linear-solver:before {
	content: '\e901';
}

.icon-model-height:before {
	content: '\e902';
}

.icon-nonlinear-solver:before {
	content: '\e903';
}

.icon-show-connections:before {
	content: '\e904';
}

.icon-show-connections-2:before {
	content: '\e904';
}

.icon-home2:before {
	content: '\e912';
}

.icon-floppy-disk:before {
	content: '\e962';
}

.icon-search:before {
	content: '\e986';
}

.icon-enlarge:before {
	content: '\e989';
}

.icon-shrink:before {
	content: '\e98a';
}

.icon-enlarge2:before {
	content: '\e98b';
}

.icon-shrink2:before {
	content: '\e98c';
}

.icon-text-height:before {
	content: '\ea5f';
}

/* End of section is copied from the icomoon style.css file that is generated by icomoon */


[class^='icon-']:before,
[class*=' icon-']:before {
	display: inline-block;
	font: normal normal normal 14px/1 fontello;
	font-size: inherit;
	text-rendering: auto;
	-webkit-font-smoothing: antialiased;
	-moz-osx-font-smoothing: grayscale;
}

/* makes the font 33%% larger relative to the icon container */
.bigger-font {
	font-size: 1.33333333em;
	line-height: 0.75em;
	vertical-align: -15%;
}

/* To get SVG and PNG images of these icon fonts, use https://icomoon.io/app/#/select */
.icon-home:before {
	content: '\e800';
}

.icon-left-big:before {
	content: '\e801';
}

.icon-right-big:before {
	content: '\e802';
}

.icon-up-big:before {
	content: '\e803';
}

.icon-resize-full:before {
	content: '\e804';
}

.icon-resize-small:before {
	content: '\e805';
}

.icon-resize-vertical:before {
	content: '\e807';
}

.icon-floppy:before {
	content: '\e808';
}

.icon-minus:before {
	content: '\e80a';
}

.icon-exchange:before {
	content: '\f0ec';
}

.icon-terminal:before {
	content: '\f120';
}

.icon-help:before {
	content: '\f128';
}

.icon-eraser:before {
	content: '\f12d';
}

.icon-sort-number-up:before {
	content: '\f162';
}

.icon-right:before {
	content: '\f178';
}

.icon-map-signs:before {
	content: '\f277';
}
